﻿using System.Buffers;
using System.IO.Pipelines;
using System.Text;
using System.Text.RegularExpressions;

namespace AdventOfCode;

public static partial class Day1
{
    [GeneratedRegex(pattern: "0", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Zero();
    
    [GeneratedRegex(pattern: "one|1", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex One();
    
    [GeneratedRegex(pattern: "two|2", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Two();
    
    [GeneratedRegex(pattern: "three|3", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Three();
    
    [GeneratedRegex(pattern: "four|4", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Four();
    
    [GeneratedRegex(pattern: "five|5", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Five();
    
    [GeneratedRegex(pattern: "six|6", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Six();
    
    [GeneratedRegex(pattern: "seven|7", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Seven();
    
    [GeneratedRegex(pattern: "eight|8", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Eight();
    
    [GeneratedRegex(pattern: "nine|9", RegexOptions.Compiled | RegexOptions.IgnoreCase)]
    private static partial Regex Nine();

    private static readonly Dictionary<Regex, int> matchers = new()
    {
        { Zero(), 0 },
        { One(), 1 },
        { Two(), 2 },
        { Three(), 3 },
        { Four(), 4 },
        { Five(), 5 },
        { Six(), 6 },
        { Seven(), 7 },
        { Eight(), 8 },
        { Nine(), 9 },
    };
    
    public class Result
    {
        public int Sum;
    }
    
    public static async Task<int> ExecuteAsync(string filePath)
    {
        if (!File.Exists(filePath))
        {
            return 0;
        }

        await using var fs = new FileStream(filePath, FileMode.Open);
        return await ExecuteAsync(fs);
    }

    public static async Task<int> ExecuteAsync(Stream s)
    {
        Result result = new();
        var pipe = new Pipe();
        var writing = FillPipeAsync(s, pipe.Writer);
        var reading = ReadPipeAsync(pipe.Reader, result);

        await Task.WhenAll(writing, reading);
        
        return result.Sum;
    }

    private static async Task FillPipeAsync(Stream fs, PipeWriter writer)
    {
        const int minBufferSize = 512;

        while (true)
        {
            var memory = writer.GetMemory(minBufferSize);
            try
            {
                var bytesRead = await fs.ReadAsync(memory, CancellationToken.None);
                if (bytesRead == 0)
                {
                    break;
                }
                writer.Advance(bytesRead);
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
                break;
            }

            var result = await writer.FlushAsync();
            if (result.IsCompleted)
            {
                break;
            }
        }

        await writer.CompleteAsync();
    }
    
    private static async Task<int> ReadPipeAsync(PipeReader reader, Result result)
    {
        var sum = 0;
        result.Sum = 0;
        while (true)
        {
            var readResult = await reader.ReadAsync();
            var buffer = readResult.Buffer;

            while (TryReadLine(ref buffer, out var line))
            {
                sum += ProcessLine(line);
            }
            
            reader.AdvanceTo(buffer.Start, buffer.End);

            if (readResult.IsCompleted)
            {
                break;
            }
        }

        await reader.CompleteAsync();
        result.Sum = sum;
        return sum;
    }

    private static bool TryReadLine(ref ReadOnlySequence<byte> buffer, out ReadOnlySequence<byte> line)
    {
        var position = buffer.PositionOf((byte)'\n');

        if (position == null)
        {
            line = default;
            return false;
        }

        line = buffer.Slice(0, position.Value);
        buffer = buffer.Slice(buffer.GetPosition(1, position.Value));
        return true;
    }
    
    private static int ProcessLine(ReadOnlySequence<byte> line)
    {
        var str = Encoding.UTF8.GetString(line);
        return string.IsNullOrWhiteSpace(str) ? 0 : ParseString(str);
    }

    /// <summary>
    /// Assuming we don't have to deal with endianness and exotic code points
    /// </summary>
    /// <param name="str"></param>
    /// <returns></returns>
    public static int ParseString(string str)
    {
        var strSpan = str.AsSpan();
        int? firstDigit = default;
        int? lastDigit = default;

        for (var index = 0; index < strSpan.Length; index++)
        {
            var span = strSpan[0 .. (index + 1)];
            if (span.IsEmpty)
            {
                continue;
            }

            foreach (var kv in matchers)
            {
                if (!kv.Key.IsMatch(span))
                {
                    continue;
                }
                
                firstDigit ??= kv.Value;
                break;
            }

            if (firstDigit != null)
            {
                break;
            }
        }

        for (var i = strSpan.Length - 1; i >= 0; i--)
        {
            var span = strSpan[i .. strSpan.Length];
            if (span.IsEmpty)
            {
                continue;
            }
            
            foreach (var kv in matchers)
            {
                if (!kv.Key.IsMatch(span))
                {
                    continue;
                }
                
                lastDigit ??= kv.Value;
                break;
            }
            
            if (lastDigit != null)
            {
                break;
            }
        }

        if (firstDigit == null)
        {
            return 0;
        }

        lastDigit ??= firstDigit;
        
        return !int.TryParse($"{firstDigit.Value}{lastDigit.Value}", out var result)
            ? 0
            : result;
    }
}